#!/usr/bin/env python3
"""
Created on Mon Jun 12 16:23:54 2023.

@author: ghiggi
"""
import re
from typing import Any

from pandas.api.extensions import (
    ExtensionDtype,
    register_extension_dtype,
)

from tstore.tsdf.ts_class import TS

#### Notes
# https://pandas.pydata.org/pandas-docs/stable/reference/extensions.html
# https://pandas.pydata.org/pandas-docs/stable/development/extending.html#extensionarray
# https://pandas.pydata.org/pandas-docs/stable/development/extending.html#extension-types
# https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.api.extensions.ExtensionArray.html
# https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.api.extensions.ExtensionDtype.html

# https://github.com/geopandas/geopandas/blob/83ab5c63890a2575b95ede6b8a8ef469753c9605/geopandas/array.py#L33
# https://github.com/geopandas/geopandas/blob/83ab5c63890a2575b95ede6b8a8ef469753c9605/geopandas/array.py#L261
# https://stackoverflow.com/questions/68893521/simple-example-of-pandas-extensionarray

# https://itnext.io/guide-to-pandas-extension-types-and-how-to-create-your-own-3b213d689c86


#### Extension DType


class TSDtype(ExtensionDtype):
    """An ExtensionDtype for TS time series data."""

    # Scalar type for the array
    type = TS

    # Required for all parameterized dtypes
    _metadata = ("ts_class",)
    _match = re.compile(r"TS\[(?P<ts_class>.+)\]")

    def __init__(self, ts_class=None):
        if ts_class is None:
            ts_class = "-"
        self._ts_class = ts_class

    def __str__(self) -> str:
        """String representation."""
        return f"TS[{self._ts_class}]"

    # Required for all ExtensionDtype subclasses
    @property
    def name(self) -> str:
        """A string representation of the dtype."""
        return str(self)

    # TestDtypeTests
    def __hash__(self) -> int:
        """Hash."""
        return hash(str(self))

    # TestDtypeTests
    def __eq__(self, other: Any) -> bool:
        """Test TS equality."""
        if isinstance(other, str):
            return self.name == other
        return isinstance(other, type(self)) and self.ts_class == other.ts_class

    # Required for pickle compat (see GH26067)
    def __setstate__(self, state) -> None:
        """__setstate__."""
        self._ts_class = state["ts_class"]

    # Required for all ExtensionDtype subclasses
    @classmethod
    def construct_array_type(cls):
        """Return the array type associated with this dtype."""
        from tstore.tsdf.extensions.array import TSArray

        return TSArray

    # Recommended for parameterized dtypes
    @classmethod
    def construct_from_string(cls, string: str):
        """
        Construct an TSDtype from a string.

        Example:
        -------
        >>> TSDtype.construct_from_string("TS[pandas]")
        TS['pandas']
        """
        if not isinstance(string, str):
            msg = f"'construct_from_string' expects a string, got {type(string)}"
            raise TypeError(msg)

        msg = f"Cannot construct a '{cls.__name__}' from '{string}'"
        match = cls._match.match(string)

        if match:
            d = match.groupdict()
            try:
                return cls(ts_class=d["ts_class"])
            except (KeyError, TypeError, ValueError) as err:
                raise TypeError(msg) from err
        else:
            raise TypeError(msg)

    @property
    def ts_class(self) -> str:
        """The TS object class."""
        return self._ts_class


register_extension_dtype(TSDtype)



#!/usr/bin/env python3
"""
Created on Mon Jun 12 23:22:02 2023.

@author: ghiggi
"""
import pandas as pd

from tstore.archive.checks import check_is_tstore, check_tstore_ids
from tstore.archive.io import define_attributes_filepath


def read_attributes(base_dir, tstore_ids=None):
    """Read TStore attributes in a pandas.DataFrame."""
    # Checks
    base_dir = check_is_tstore(base_dir)
    tstore_ids = check_tstore_ids(tstore_ids, base_dir=base_dir)
    # Retrieve filepath
    fpath = define_attributes_filepath(base_dir)
    # Read TStore attributes
    df = pd.read_parquet(fpath, engine="pyarrow", dtype_backend="pyarrow")
    df.index = df.index.astype("string[pyarrow]")
    # Subset tstore_ids
    # TODO: maybe possible to pass filters to pd.read_parquet
    return df



#!/usr/bin/env python3
"""
Created on Mon Jun 12 17:57:40 2023.

@author: ghiggi
"""
from tstore.tslong.pyarrow import open_tslong as pyarrow_reader


def open_tslong(
    base_dir,
    ts_variables=None,
    start_time=None,
    end_time=None,
    tstore_ids=None,
    columns=None,
    filesystem=None,
    use_threads=True,
):
    """Open TStore into long-format into pandas.DataFrame."""
    # Read exploiting pyarrow
    # - We use pyarrow to avoid pandas copies at concatenation and join operations !
    tslong_pyarrow = pyarrow_reader(
        base_dir,
        ts_variables=ts_variables,
        start_time=start_time,
        end_time=end_time,
        tstore_ids=tstore_ids,
        columns=columns,
        filesystem=filesystem,
        use_threads=use_threads,
    )

    # Conversion to pandas
    tslong_pandas = tslong_pyarrow.to_pandas()
    return tslong_pandas



"""TStore I/O tools."""

import glob
import os
import shutil


def check_tstore_structure(tstore_structure):
    """Check validity of TStore structure."""
    if tstore_structure not in ["id-var", "var-id"]:
        raise ValueError("Valid tstore_structure are 'id-var' and 'var-id'.")
    return tstore_structure


def check_tstore_directory(base_dir, overwrite):
    """Check TStore directory."""
    # Check if exists, remove or raise error
    if os.path.exists(base_dir):
        if overwrite:
            shutil.rmtree(base_dir)
        else:
            raise ValueError("TStore already existing at {base_dir}")
    # Create directory
    os.makedirs(base_dir, exist_ok=True)
    return base_dir


def define_attributes_filepath(base_dir):
    """Define filepath of TStore attributes."""
    fpath = os.path.join(base_dir, "_attributes.parquet")
    return fpath


def define_tsarray_filepath(base_dir, tstore_id, ts_variable, tstore_structure):
    """Define filepath of a TStore TS."""
    if tstore_structure == "id-var":
        fpath = os.path.join(base_dir, tstore_id, ts_variable)
    elif tstore_structure == "var-id":
        fpath = os.path.join(base_dir, ts_variable, tstore_id)
    else:
        raise ValueError("Valid tstore_structure are 'id-var' and 'var-id'.")
    return fpath


def define_metadata_filepath(base_dir):
    """Define TStore metadata filepath."""
    metadata_fpath = os.path.join(base_dir, "tstore_metadata.yml")
    return metadata_fpath


def get_tstore_structure(base_dir):
    """Get TStore structure."""
    from tstore.archive.readers import read_metadata

    metadata = read_metadata(base_dir=base_dir)
    return metadata["tstore_structure"]


def get_ts_info(base_dir, ts_variable):
    """Retrieve filepaths and tstore_ids for a specific ts_variable."""
    tstore_structure = get_tstore_structure(base_dir)
    if tstore_structure == "id-var":
        fpaths = glob.glob(os.path.join(base_dir, "*", ts_variable))
        tstore_ids = [os.path.basename(os.path.dirname(fpath)) for fpath in fpaths]
    elif tstore_structure == "var-id":
        fpaths = glob.glob(os.path.join(base_dir, ts_variable, "*"))
        tstore_ids = [os.path.basename(fpath) for fpath in fpaths]
    else:
        raise ValueError("Valid tstore_structure are 'id-var' and 'var-id'.")
    return fpaths, tstore_ids



#!/usr/bin/env python3
"""
Created on Mon Jun 12 22:21:47 2023.

@author: ghiggi
"""


def get_dataset_partitioning_columns(fpath):
    """Infer partitioning columns from dataset file path."""
    # TODO: implement logic for that
    partitioning_columns = ["year", "month"]
    return partitioning_columns



#!/usr/bin/env python3
"""
Created on Mon Jun 12 15:35:06 2023.

@author: ghiggi
"""
import yaml

from tstore.archive.io import define_metadata_filepath


def _read_yaml_metadata(fpath):
    """Read metadata YAML file."""
    with open(fpath) as file:
        metadata = yaml.safe_load(file)
    return metadata


def read_metadata(base_dir):
    """Read TStore metadata."""
    metadata_fpath = define_metadata_filepath(base_dir)
    metadata = _read_yaml_metadata(metadata_fpath)
    return metadata



#!/usr/bin/env python3
"""
Created on Mon Jun 12 23:22:16 2023.

@author: ghiggi
"""
import pyarrow.parquet as pq

from tstore.archive.checks import check_is_tstore, check_tstore_ids
from tstore.archive.io import define_attributes_filepath


def get_tstore_ids_filters(tstore_ids=None):
    """Define filters for Parquet Dataset subsetting at read-time."""
    # TODO implement logic
    filters = None
    return filters


def read_attributes(base_dir, tstore_ids=None, filesystem=None, use_threads=True):
    """Read TStore attributes in a pyarrow.Table."""
    # Checks
    base_dir = check_is_tstore(base_dir)
    tstore_ids = check_tstore_ids(tstore_ids, base_dir=base_dir)
    # Define filters
    filters = get_tstore_ids_filters(tstore_ids=tstore_ids)
    # Retrieve filepath
    fpath = define_attributes_filepath(base_dir)
    # Read TStore attributes
    table = pq.read_table(
        fpath,
        use_pandas_metadata=True,
        filters=filters,
        filesystem=filesystem,
        use_threads=use_threads,
    )
    return table



#!/usr/bin/env python3
"""
Created on Mon Jun 12 22:24:07 2023.

@author: ghiggi
"""
import dask.dataframe as dd
import pandas as pd

from tstore.archive.ts.filtering import get_time_filters
from tstore.archive.ts.partitioning import get_dataset_partitioning_columns


def open_ts(
    fpath,
    columns=None,
    start_time=None,
    end_time=None,
    split_row_groups=False,
    # Dask-specific
    calculate_divisions=True,
    ignore_metadata_file=False,
    **kwargs,
):
    """Open a time series into a Dask.DataFrame."""
    # Define filters argument
    filters = get_time_filters(start_time=start_time, end_time=end_time)

    # Define Apache Arrow settings
    arrow_to_pandas = {
        "zero_copy_only": False,  # Default is False. If True, raise error if doing copies
        "strings_to_categorical": False,
        "date_as_object": False,  # Default is True. If False convert to datetime64[ns]
        "timestamp_as_object": False,  # Default is True. If False convert to np.datetime64[ns]
        "use_threads": True,  #  parallelize the conversion using multiple threads.
        "safe": True,
        "split_blocks": False,
        "ignore_metadata": False,  # Default False. If False, use the 'pandas' metadata to get the Index
        "types_mapper": pd.ArrowDtype,  # Ensure pandas is created with Arrow dtype
    }

    # Read Apache Parquet
    df = dd.read_parquet(
        fpath,
        engine="pyarrow",
        dtype_backend="pyarrow",
        index=None,  # None --> Read back original time-index
        # Filtering
        columns=columns,  # Specify columns to load
        filters=filters,  # Row-filtering at read-time
        # Metadata options
        calculate_divisions=calculate_divisions,  # Calculate divisions from metadata
        ignore_metadata_file=ignore_metadata_file,  # True can slowdown a lot reading
        # Partitioning
        split_row_groups=split_row_groups,  # False --> Each file a partition
        # Arrow options
        arrow_to_pandas=arrow_to_pandas,
        # Other options,
        **kwargs,
    )

    # Drop partitioning columns
    partitioning_columns = get_dataset_partitioning_columns(fpath)
    df = df.drop(columns=partitioning_columns)

    return df



"""TStore writers."""

import yaml

from tstore.archive.io import (
    define_attributes_filepath,
    define_metadata_filepath,
)


def _write_yaml_metadata(metadata, fpath):
    """Write metadata YAML file."""
    with open(fpath, "w") as file:
        yaml.dump(metadata, file)


def write_attributes(df, base_dir):
    """Write static attributes dataframe.

    Assume df pandas !
    TODO: polars, pyarrow does not have index
    --> Ensure tstore_id column
    """
    fpath = define_attributes_filepath(base_dir)
    df.index = df.index.astype("string[pyarrow]")
    df.to_parquet(
        fpath,
        engine="pyarrow",
        compression="snappy",
        index=True,
    )


def write_metadata(base_dir, ts_variables, tstore_structure):
    """Write TStore metadata file."""
    metadata_fpath = define_metadata_filepath(base_dir)
    metadata = {}
    metadata["ts_variables"] = ts_variables
    metadata["tstore_structure"] = tstore_structure
    _write_yaml_metadata(metadata=metadata, fpath=metadata_fpath)



"""TSDF."""

import pandas as pd


class TSDF(pd.DataFrame):
    """A dataframe class with additional functionality for TSArray data."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def to_tstore(
        self,
        base_dir,
        partition_str=None,
        tstore_structure="id-var",
        overwrite=True,
    ):
        """Write TStore from TSDF object."""
        from tstore.tsdf.writer import write_tstore

        _ = write_tstore(
            self,
            base_dir=base_dir,
            partition_str=partition_str,
            tstore_structure=tstore_structure,
            overwrite=overwrite,
        )



#!/usr/bin/env python3
"""
Created on Mon Jun 12 21:32:37 2023.

@author: ghiggi
"""
import datetime

import numpy as np


def get_available_ts_variables(base_dir):
    """Get available TStore timeseries."""
    from tstore.archive.readers import read_metadata

    metadata = read_metadata(base_dir=base_dir)
    return metadata["ts_variables"]


def check_is_tstore(base_dir):
    """Check is a TStore."""
    # TODO
    pass
    return base_dir


def check_ts_variables(ts_variables, base_dir):
    """Check valid ts_variables.

    If None, return all the available ts_variables.
    """
    available_ts_variables = get_available_ts_variables(base_dir)
    if isinstance(ts_variables, str):
        ts_variables = [ts_variables]
    if isinstance(ts_variables, type(None)):
        ts_variables = available_ts_variables
    ts_variables = np.array(ts_variables)
    unvalid_ts_variables = ts_variables[np.isin(ts_variables, available_ts_variables, invert=True)]
    ts_variables = ts_variables.tolist()
    unvalid_ts_variables = unvalid_ts_variables.tolist()
    if len(unvalid_ts_variables) > 0:
        raise ValueError(
            f"Valid ts_variables are {ts_variables}. Invalid: {unvalid_ts_variables}.",
        )
    return ts_variables


def check_tstore_ids(tstore_ids, base_dir):
    """Check valid tstore_ids.

    If tstore_ids=None, return None.
    """
    # TODO:
    pass
    return tstore_ids


def check_time(time):
    """Check time validity.

    It returns a datetime.datetime object.

    Parameters
    ----------
    time : (datetime.datetime, datetime.date, np.datetime64, str)
        Time object.
        Accepted types: datetime.datetime, datetime.date, np.datetime64, str
        If string type, it expects the isoformat 'YYYY-MM-DD hh:mm:ss'.

    Returns
    -------
    time : datetime.datetime
        datetime.datetime object.

    """
    # TODO: adapt to return the more appropriate object !!!
    if not isinstance(
        time,
        (datetime.datetime, datetime.date, np.datetime64, np.ndarray, str),
    ):
        raise TypeError(
            "Specify time with datetime.datetime objects or a " "string of format 'YYYY-MM-DD hh:mm:ss'.",
        )
    # If numpy array with datetime64 (and size=1)
    if isinstance(time, np.ndarray):
        if np.issubdtype(time.dtype, np.datetime64):
            if time.size == 1:
                time = time.astype("datetime64[ns]").tolist()
            else:
                raise ValueError("Expecting a single timestep!")
        else:
            raise ValueError("The numpy array does not have a np.datetime64 dtype!")
    # If np.datetime64, convert to datetime.datetime
    if isinstance(time, np.datetime64):
        time = time.astype("datetime64[s]").tolist()
    # If datetime.date, convert to datetime.datetime
    if not isinstance(time, (datetime.datetime, str)):
        time = datetime.datetime(time.year, time.month, time.day, 0, 0, 0)
    if isinstance(time, str):
        try:
            time = datetime.datetime.fromisoformat(time)
        except ValueError:
            raise ValueError("The time string must have format 'YYYY-MM-DD hh:mm:ss'")
    return time


def check_start_end_time(start_time, end_time):
    """Check start_time and end_time validity."""
    # Format time input
    if start_time is not None:
        start_time = check_time(start_time)
    if end_time is not None:
        end_time = check_time(end_time)
    # Check start_time and end_time are chronological
    if start_time is not None and end_time is not None and start_time > end_time:
        raise ValueError("Provide start_time occurring before of end_time")
    return (start_time, end_time)



#!/usr/bin/env python3
"""
Created on Mon Jun 12 15:49:54 2023.

@author: ghiggi
"""
import numpy as np

from tstore.archive.io import (
    check_tstore_directory,
    check_tstore_structure,
    define_tsarray_filepath,
)
from tstore.tsdf.extensions.ts_dtype import TSDtype


def _get_ts_variables(df):
    """Get list of TSArray columns."""
    columns = np.array(list(df.columns))
    ts_variables = columns[[isinstance(df[column].dtype, TSDtype) for column in columns]].tolist()
    return ts_variables


def _get_static_columns(df):
    """Get list of non-TSArray columns."""
    columns = list(df.columns)
    ts_variables = _get_ts_variables(df)
    static_columns = list(set(columns).symmetric_difference(ts_variables))
    return static_columns


def _write_attributes(df, base_dir):
    """Write TSDF static attributes."""
    from tstore.archive.writers import write_attributes

    static_columns = _get_static_columns(df)
    df_attributes = df[static_columns]
    df_attributes.index.name = "tstore_id"
    write_attributes(df=df_attributes, base_dir=base_dir)


def _write_ts_series(ts_series, base_dir, tstore_structure):
    """Write TSDF TSArray."""
    ts_variable = ts_series.name
    tstore_ids = ts_series.index.array.astype(str)
    for tstore_id, ts in zip(tstore_ids, ts_series):
        if ts:  # not empty
            ts_fpath = define_tsarray_filepath(
                base_dir=base_dir,
                tstore_id=tstore_id,
                ts_variable=ts_variable,
                tstore_structure=tstore_structure,
            )
            ts.to_disk(ts_fpath)


def _write_tsarrays(df, base_dir, tstore_structure):
    """Write TSDF TSArrays."""
    tsarray_columns = _get_ts_variables(df)
    for column in tsarray_columns:
        _write_ts_series(
            ts_series=df[column],
            base_dir=base_dir,
            tstore_structure=tstore_structure,
        )


def _write_metadata(df, base_dir, tstore_structure):
    """Write TStore metadata."""
    from tstore.archive.writers import write_metadata

    ts_variables = _get_ts_variables(df)
    write_metadata(
        base_dir=base_dir,
        ts_variables=ts_variables,
        tstore_structure=tstore_structure,
    )


def write_tstore(
    df,
    base_dir,
    partition_str=None,
    tstore_structure="id-var",
    overwrite=True,
):
    """Write TStore from TSDF object."""
    # Checks
    tstore_structure = check_tstore_structure(tstore_structure)
    base_dir = check_tstore_directory(base_dir, overwrite=overwrite)

    # Write static attributes
    _write_attributes(df, base_dir=base_dir)

    # Write TSArrays
    _write_tsarrays(df, base_dir=base_dir, tstore_structure=tstore_structure)

    # Write TSArrays metadata
    _write_metadata(df, base_dir=base_dir, tstore_structure=tstore_structure)



#!/usr/bin/env python3
"""
Created on Sun Jun 11 22:47:54 2023.

@author: ghiggi
"""

import dask.dataframe as dd
import numpy as np
import pandas as pd
from pandas.api.extensions import ExtensionArray

from tstore.tsdf.extensions.ts_dtype import TSDtype
from tstore.tsdf.ts_class import TS

#### Notes
# https://pandas.pydata.org/pandas-docs/stable/reference/extensions.html
# https://pandas.pydata.org/pandas-docs/stable/development/extending.html#extensionarray
# https://pandas.pydata.org/pandas-docs/stable/development/extending.html#extension-types
# https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.api.extensions.ExtensionArray.html
# https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.api.extensions.ExtensionDtype.html

# https://github.com/geopandas/geopandas/blob/83ab5c63890a2575b95ede6b8a8ef469753c9605/geopandas/array.py#L33
# https://github.com/geopandas/geopandas/blob/83ab5c63890a2575b95ede6b8a8ef469753c9605/geopandas/array.py#L261
# https://stackoverflow.com/questions/68893521/simple-example-of-pandas-extensionarray

# https://itnext.io/guide-to-pandas-extension-types-and-how-to-create-your-own-3b213d689c86


####--------------------------------------------------------------------------.
def get_tabular_object_type(obj):
    """Get inner class of the TS object."""
    if isinstance(obj, dd.DataFrame):
        return "dask.DataFrame"
    if isinstance(obj, pd.DataFrame):
        return "dask.Series"
    if isinstance(obj, pd.DataFrame):
        return "pandas.DataFrame"
    if isinstance(obj, pd.Series):
        return "pandas.Series"
    return type(obj).__name__


class TSArray(ExtensionArray):
    """An ExtensionArray for TS objects, holding the array-based implementations."""

    _dtype = TSDtype()

    def __init__(self, data, copy: bool = False):
        self._data = np.array(data, copy=copy)

    @property
    def _class(self):
        """Define inner TS class."""
        # Infer TS class from TS objects
        # TODO: Now first. In future loop and check unique ...

        #### BUG HERE
        # If no matching index, when joining during open_tsdf, we get
        #  array([nan, nan, nan, nan], dtype=object) here
        # --> Need to create empty object !
        # TS[empty]
        ts_object = self._data[0]
        tabular_object = getattr(ts_object, "data", pd.Series())
        ts_class = get_tabular_object_type(tabular_object)
        return ts_class

    def __str__(self):
        """String representation."""
        # TODO print
        return str(self._data)

    def __repr__(self):
        """Repr representation."""
        n = len(self._data)
        return f"TSArray composed of {n} TS objects."

    # Required for all ExtensionArray subclasses
    def __getitem__(self, index: int):
        """Select a subset of self."""
        if isinstance(index, int):
            return self._data[index]
        # Check index for TestGetitemTests
        index = pd.core.indexers.check_array_indexer(self, index)
        return type(self)(self._data[index])

    # TestSetitemTests
    def __setitem__(self, index: int, value: TS) -> None:
        """Set one or more values in-place."""
        # Check index for TestSetitemTests
        index = pd.core.indexers.check_array_indexer(self, index)

        # Upcast to value's type (if needed) for TestMethodsTests
        if self._data.dtype < type(value):
            self._data = self._data.astype(type(value))

        # TODO: Validate value for TestSetitemTests
        # value = self._validate_setitem_value(value)

        self._data[index] = value

    # Required for all ExtensionArray subclasses
    def __len__(self) -> int:
        """Length of this array."""
        return len(self._data)

    # Required for all ExtensionArray subclasses
    @pd.core.ops.unpack_zerodim_and_defer("__eq__")
    def __eq__(self, other):
        """Equality behaviour."""
        # TODO: how to compare list of TS objects
        return False

    # Required for all ExtensionArray subclasses
    @classmethod
    def _from_sequence(cls, data, dtype=None, copy: bool = False):
        """Construct a new TSArray from a sequence of TS."""
        if dtype is None:
            dtype = TSDtype()

        if not isinstance(dtype, TSDtype):
            msg = f"'{cls.__name__}' only supports 'TSDtype' dtype"
            raise ValueError(msg)
        return cls(data, copy=copy)

    # Required for all ExtensionArray subclasses
    @classmethod
    def _concat_same_type(cls, to_concat):
        """Concatenate multiple TSArrays."""
        # Ensure same TS class
        counts = pd.value_counts([array.dtype.ts_class for array in to_concat])
        if counts.size > 1:
            raise ValueError("The TS objects must all be of the same type.")

        return cls(np.concatenate(to_concat))

    # Required for all ExtensionArray subclasses
    @property
    def dtype(self):
        """An instance of TSDtype."""
        return TSDtype(self._class)

    # Required for all ExtensionArray subclasses
    @property
    def nbytes(self) -> int:
        """The number of bytes needed to store this object in memory."""
        return self._data.nbytes

    @property
    def ts_class(self):
        """TS inner class."""
        return self.dtype.ts_class

    # Required for all ExtensionArray subclasses
    def isna(self):
        """A 1-D array indicating if the TS is missing."""
        return pd.isnull(self._data)

    # Required for all ExtensionArray subclasses
    def copy(self):
        """Return a copy of the array."""
        copied = self._data.copy()
        return type(self)(copied)

    # Required for all ExtensionArray subclasses
    def take(self, indices, allow_fill=False, fill_value=None):
        """Take elements from an array."""
        if allow_fill and fill_value is None:
            fill_value = self.dtype.na_value

        result = pd.core.algorithms.take(
            self._data,
            indices,
            allow_fill=allow_fill,
            fill_value=fill_value,
        )
        return self._from_sequence(result)



#!/usr/bin/env python3
"""
Created on Mon Jun 12 22:27:07 2023.

@author: ghiggi
"""


def get_time_filters(start_time=None, end_time=None):
    """Define filters for Parquet Dataset subsetting at read-time."""
    # TODO implement logic
    filters = None
    return filters



#!/usr/bin/env python3
"""
Created on Mon Jun 12 22:19:51 2023.

@author: ghiggi
"""
import pyarrow.parquet as pq

from tstore.archive.ts.filtering import get_time_filters
from tstore.archive.ts.partitioning import get_dataset_partitioning_columns


def open_ts(
    fpath,
    start_time=None,
    end_time=None,
    columns=None,
    split_row_groups=False,
    # pyarrow-specific
    filesystem=None,
    use_threads=True,
):
    """Open a TS into a pyarrow.Table."""
    # Define filters argument
    filters = get_time_filters(start_time=start_time, end_time=end_time)

    # Read Option 1
    # - https://arrow.apache.org/docs/python/generated/pyarrow.parquet.read_table.html#pyarrow-parquet-read-table
    table = pq.read_table(
        fpath,
        use_pandas_metadata=True,
        columns=columns,
        filters=filters,
        filesystem=filesystem,
        use_threads=use_threads,
    )
    # Read Option 2
    # - https://arrow.apache.org/docs/python/generated/pyarrow.parquet.ParquetDataset.html
    # Create a ParquetDataset object
    # dataset = pq.ParquetDataset(fpath,
    #                             filters=filters,
    #                             filesystem=filesystem,
    #                             # Specific to ParquetDataset
    #                             # metadata_nthreads=1,
    #                             split_row_groups=split_row_groups,

    # )
    # table = dataset.read(columns=columns,
    #                       use_pandas_metadata=True,
    #                       use_threads=use_threads,
    #
    # )

    # Remove partitioning columns
    partitioning_columns = get_dataset_partitioning_columns(fpath)
    table = table.drop(partitioning_columns)
    return table



"""tstore."""

import contextlib
from importlib.metadata import PackageNotFoundError, version

from tstore.tsdf import TSDF
from tstore.tsdf.extensions.array import TSArray
from tstore.tsdf.extensions.ts_dtype import TSDtype
from tstore.tsdf.reader import open_tsdf
from tstore.tsdf.ts_class import TS
from tstore.tslong.pandas import open_tslong

__all__ = [
    "open_tsdf",
    "TSArray",
    "TSDtype",
    "TS",
    "TSDF",
]

# Get version
with contextlib.suppress(PackageNotFoundError):
    __version__ = version("tstore")



#!/usr/bin/env python3
"""
Created on Mon Jun 12 17:57:14 2023.

@author: ghiggi
"""
from functools import reduce

import numpy as np
import pyarrow as pa

from tstore.archive.attributes.pyarrow import read_attributes
from tstore.archive.checks import (
    check_is_tstore,
    check_start_end_time,
    check_ts_variables,
    check_tstore_ids,
)
from tstore.archive.io import get_ts_info
from tstore.archive.ts.pyarrow import open_ts


def _read_ts(
    fpath,
    tstore_id,
    start_time=None,
    end_time=None,
    columns=None,
    filesystem=None,
    use_threads=True,
):
    """Read a single TS in pyarrow long-format."""
    # Read TS in pyarrow long format
    table = open_ts(
        fpath=fpath,
        start_time=start_time,
        end_time=end_time,
        columns=columns,
        filesystem=filesystem,
        use_threads=use_threads,
    )
    # Add tstore_id
    tstore_id = str(tstore_id)
    tstore_id = pa.array(np.repeat(tstore_id, len(table)), type=pa.string())
    table = table.add_column(0, "tstore_id", tstore_id)
    return table


def _read_ts_variable(
    base_dir,
    ts_variable,
    start_time=None,
    end_time=None,
    columns=None,
    filesystem=None,
    use_threads=True,
):
    """Read a TStore ts_variable into pyarrow long-format."""
    # Find TS and associated TStore IDs
    fpaths, tstore_ids = get_ts_info(base_dir=base_dir, ts_variable=ts_variable)
    # Read each TS
    list_tables = [
        _read_ts(
            fpath=fpath,
            tstore_id=tstore_id,
            start_time=start_time,
            end_time=end_time,
            columns=columns,
            filesystem=filesystem,
            use_threads=use_threads,
        )
        for fpath, tstore_id in zip(fpaths, tstore_ids)
    ]
    # Concatenate the tables
    table = pa.concat_tables(list_tables)
    return table


def _join_tables(left_table, right_table):
    """Joining functions of pyarrow tables."""
    # TODO: update keys to 'time'
    return left_table.join(
        right_table,
        keys=["timestamp", "tstore_id"],
        join_type="full outer",
    )


def _read_ts_variables(
    base_dir,
    ts_variables,
    start_time=None,
    end_time=None,
    columns=None,
    filesystem=None,
    use_threads=True,
):
    """Read TStore ts_variables into pyarrow long-format."""
    # TODO: columns must be a dictionary {ts_variable: [...]}

    # Read TS of all ts_variables in long-format
    list_tables = [
        _read_ts_variable(
            base_dir=base_dir,
            ts_variable=ts_variable,
            start_time=start_time,
            end_time=end_time,
            columns=columns,  # columns[ts_variable] in future
            filesystem=filesystem,
            use_threads=use_threads,
        )
        for ts_variable in ts_variables
    ]

    # Check that each table has different column names
    # --> Except from 'time' and 'tstore_id' on which align !
    # TODO:

    # Check each table has 'time' and 'tstore_id'
    # TODO:

    # Iteratively join the tables
    table = reduce(_join_tables, list_tables)

    return table


def open_tslong(
    base_dir,
    ts_variables=None,
    start_time=None,
    end_time=None,
    tstore_ids=None,
    columns=None,
    filesystem=None,
    use_threads=True,
):
    """Open TStore into long-format into pyarrow.Table."""
    # TODO: columns must be a dictionary {ts_variable: [...]}
    # --> Preprocess in dictionary {ts_variable: None ... }

    # Checks
    base_dir = check_is_tstore(base_dir)
    ts_variables = check_ts_variables(ts_variables, base_dir=base_dir)
    tstore_ids = check_tstore_ids(tstore_ids, base_dir=base_dir)
    start_time, end_time = check_start_end_time(start_time, end_time)

    # Get list of tslong for each ts_variable
    table = _read_ts_variables(
        base_dir=base_dir,
        ts_variables=ts_variables,
        start_time=start_time,
        end_time=end_time,
        columns=columns,
        filesystem=filesystem,
        use_threads=use_threads,
    )

    # Read TStore attributes
    table_attrs = read_attributes(
        base_dir=base_dir,
        tstore_ids=tstore_ids,
        filesystem=filesystem,
        use_threads=use_threads,
    )
    # Join (duplicate) table_attrs on table
    tslong = table.join(table_attrs, keys=["tstore_id"], join_type="full outer")

    # Return tslong
    # TODO: add class ?
    return tslong



#!/usr/bin/env python3
"""
Created on Mon Jun 12 22:17:59 2023.

@author: ghiggi
"""

import dask.dataframe as dd
import pandas as pd


def check_time_index(df):
    """Check pandas/dask index is datetime."""
    return pd.api.types.is_datetime64_any_dtype(df.index.dtype)


def add_partitioning_columns(df, partitioning_str):
    """Add partitioning columns to the dataframe."""
    # TODO: as function of TS_partitioning_string (YYYY/MM/DD) or (YY/DOY/HH)
    # dayofyear, dayofweek, hour,  minute, ...
    df["month"] = df.index.dt.month.values
    df["year"] = df.index.dt.year.values
    partition_on = ["year", "month"]
    return df, partition_on


def ensure_is_dask_dataframe(data):
    """Convert a table to a dask dataframe."""
    # Ensure object is a dask dataframe
    # - Dask Series does not have to_parquet method
    # TODO: generalize also for other object ... pandas, polars ...
    if isinstance(data, dd.Series):
        data = data.to_frame()
    return data


class TS:
    """TS object."""

    def __init__(self, data):
        """Initialize TS class."""
        # Set index as 'time'
        # --> TODO: this to adapt for polars, pyarrow ...
        data.index.name = "time"

        self.data = data

    def from_file(
        fpath,
        columns=None,
        start_time=None,
        end_time=None,
        split_row_groups=False,
        # TS class options (here dask.dataframe)
        calculate_divisions=True,
        ignore_metadata_file=False,
        **kwargs,
    ):
        """Read a time series from disk into a Dask.DataFrame."""
        # TODO: generalize to pyarrow, pandas, dask, polars
        from tstore.archive.ts.dask import open_ts

        df = open_ts(
            fpath,
            columns=columns,
            start_time=start_time,
            end_time=end_time,
            split_row_groups=split_row_groups,
            # Dask options
            calculate_divisions=calculate_divisions,
            ignore_metadata_file=ignore_metadata_file,
            **kwargs,
        )

        # Create the TS object
        return TS(df)

    def to_disk(self, fpath, partitioning_str=None):
        """Write TS object to disk."""
        # Ensure is a dask dataframe
        df = ensure_is_dask_dataframe(self.data)

        # Check the index is datetime
        check_time_index(df)

        # Check known divisions ?
        # - TODO: behaviour to determine
        if df.known_divisions:
            pass

        # Check time is sorted ?

        # Add partition columns
        df, partition_on = add_partitioning_columns(df, partitioning_str)

        # Write to Parquet
        df.to_parquet(
            fpath,
            engine="pyarrow",
            # Index option
            write_index=True,
            # Metadata
            custom_metadata=None,
            write_metadata_file=True,  # enable writing the _metadata file
            # File structure
            name_function=lambda i: f"part.{i}.parquet",  # default naming scheme
            partition_on=partition_on,
            # Encoding
            schema="infer",
            compression="snappy",
            # Writing options
            append=False,
            overwrite=False,
            ignore_divisions=False,
            compute=True,
        )



#!/usr/bin/env python3
"""
Created on Mon Jun 12 15:48:39 2023.

@author: ghiggi
"""
import pandas as pd

from tstore.archive.io import get_ts_info
from tstore.archive.readers import read_metadata
from tstore.tsdf import TSDF
from tstore.tsdf.extensions.array import TSArray
from tstore.tsdf.ts_class import TS


def _read_tsarray(base_dir, ts_variable):
    """Read a TSArray into a pd.Series."""
    # Retrieve TS fpaths and associated tstore_ids
    ts_fpaths, tstore_ids = get_ts_info(base_dir=base_dir, ts_variable=ts_variable)
    # Read TS objects
    # TODO: add option for TS format (dask, pandas, ...)
    list_ts = [TS.from_file(fpath) for fpath in ts_fpaths]

    # Create the TSArray Series
    tstore_ids = pd.Index(tstore_ids, dtype="string[pyarrow]", name="tstore_id")  # TODO: generalize
    ts_series = pd.Series(TSArray(list_ts), index=tstore_ids, name=ts_variable)
    return ts_series


def _read_tsarrays(base_dir, metadata):
    """Read list of TSArrays."""
    ts_variables = metadata["ts_variables"]
    list_ts_series = [_read_tsarray(base_dir=base_dir, ts_variable=ts_variable) for ts_variable in ts_variables]
    return list_ts_series


def open_tsdf(base_dir):
    """Open TStore into TSDF object."""
    # TODO: enable specify subset of TSArrays, attribute columns and rows to load
    # TODO: read_attributes using geopandas --> geoparquet
    # TODO: separate TSDF class if geoparquet (TSDF inherit from geopandas.GeoDataFrame ?)
    from tstore.archive.attributes.pandas import read_attributes

    # Read TStore metadata
    metadata = read_metadata(base_dir=base_dir)

    # Read TStore attributes
    df = read_attributes(base_dir)

    # Get list of TSArrays
    list_ts_series = _read_tsarrays(base_dir, metadata)

    # Join TSArrays to dataframe
    for ts_series in list_ts_series:
        df = df.join(ts_series, how="left")
        #  pd.merge(df_attrs, df_series, left_index=True, right_index=True)

    # Return the TSDF
    return TSDF(df)
